//-< Database.java >-------------------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:      1-Oct-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 16-Nov-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Class representing database for client
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
	
    /// <summary> Class representing GOODS database connection 
    /// </summary>
    public class Database
    {
        /// <summary> Set isolation level
        /// </summary>
		public virtual int IsolationLevel
        {
            set
            {
                CacheManager.isolationLevel = value;
            }
			
        }

        /// <summary> Get number of storages in the database
        /// </summary>
        public virtual CacheManager CacheManager
        {
            get
            {
                return cacheManager != null?cacheManager:CacheManager.GetCacheManager();
            }
			
        }

        public virtual int NumberOfStorages
        {
            get
            {
                return storages.Length;
            }
			
        }

        /// <summary>Get/Set root object of the first storage in the database
        /// </summary>
        public virtual Object Root
        {
            get
            {
                return storages[0].Root;
            }

            set
            {
                storages[0].Root = (Persistent) value;
            }
			
        }

        /// 
        /// <summary> Set buffer size for bulk alloc. Bulk alloc alllows
        /// to minimize number of messages sent o the server by preallocating 
        /// several OIDs during one request. 
        /// </summary>
        public virtual int AllocBufferSize
        {
            set
            {
                allocBufSize = value;
            }
			
        }
        protected internal Storage[] storages;
        protected internal IdentityHashtable notifications;
        protected internal CacheManager cacheManager;
        protected internal int allocBufSize;
		
        /// <summary> Constructor of the database
        /// </summary>
        public Database()
        {
            notifications = new IdentityHashtable();
        }
		
        /// <summary> Factory of raw binary objects which is used to pack/unpack raw binary objects
        /// </summary>
        public static RawBinaryFactory rawBinaryFactory;
		
        protected internal CondEvent onTransactionAbortEvent;
		
        /// <summary> Attach current thread to the database.
        /// This method is used in per-thread-transaction model associate
        /// the current thread with cache manager
        /// </summary>
        public virtual void  attach()
        {
            if (cacheManager == null)
            {
                cacheManager = new CacheManager();
            }
            cacheManager.attach();
        }
		
        /// <summary> Detach current thread fro the database
        /// </summary>
        public virtual void  detach()
        {
            cacheManager.detach();
            //cacheManager = null;
        }
		
        /// <summary> Transaction isolation levels.
        /// </summary>
        public const int PER_PROCESS_TRANSACTION = 0;
        public const int PER_THREAD_TRANSACTION = 1;
		
        /// <summary> Get database cache manager. In per-thread transaction model each 
        /// database has its own cache manager, providing isolation of the threads 
        /// from each other. In cooperative model all databases share the same cache
        /// managers.
        /// </summary>
		
        protected internal static System.Collections.Hashtable connections = new System.Collections.Hashtable();
		
        /// <summary> Create or reuse existed connection from connection pool
        /// </summary>
        static public Database getDatabase(String cfgFileName)
        {
            lock(typeof(Goods.Database))
            {
                System.Threading.Thread t = System.Threading.Thread.CurrentThread;
                Database db = (Database) connections[t];
                if (db == null)
                {
                    closeAllDead();
                    db = new Database();
                    if (!db.Open(cfgFileName))
                    {
                        return null;
                    }
                    connections[t] = db;
                    db.attach();
                }
                return db;
            }
        }
		
        /// <summary> Close all database active connections.
        /// </summary>
        static public void  closeAll()
        {
            lock(typeof(Goods.Database))
            {
                foreach (System.Threading.Thread t in connections.Keys) {
                    Database db = (Database) connections[t];
                    if (db.cacheManager != null)
                    {
                        db.cacheManager.detach(t);
                        db.cacheManager = null;
                    }
                    db.Close();
                }
                connections.Clear();
            }
        }
		
        /// <summary> Iterate through the listy of all active conenction and close
        /// ones associated with dead threads. 
        /// </summary>
        static public void  closeAllDead()
        {
            lock(typeof(Goods.Database))
            {
                System.Collections.IEnumerator enum_Renamed = (System.Collections.IEnumerator) connections.Keys;
                foreach (System.Threading.Thread t in connections.Keys) 
                {
                    if (!t.IsAlive)
                    {
                        Database db = (Database) connections[t];
                        if (db.cacheManager != null)
                        {
                            db.cacheManager.detach(t);
                            db.cacheManager = null;
                        }
                        db.Close();
                        connections.Remove(t);
                    }
                }
            }
        }
		
        /// <summary> Open the database
        /// </summary>
        /// <param name="cfgFileName">path to the configuration file
        /// 
        /// </param>
        public virtual bool Open(String cfgFileName)
        {
            try 
            {
                System.IO.TextReader reader = new System.IO.StreamReader(new System.IO.FileStream(cfgFileName, System.IO.FileMode.Open, System.IO.FileAccess.Read));
                try 
                { 
                    String ln = reader.ReadLine();
                    int nStorages = Int32.Parse(ln);
                    storages = new Storage[nStorages];

                    while (--nStorages >= 0)
                    {
                        ln = reader.ReadLine();
                        int i = 0;
                        for (i = 0; ln[i] != ':'; i++); 
                        int sid = Int32.Parse(ln.Substring(0, i));
                        if (sid < storages.Length)
                        {
                            if (storages[sid] != null)
                            {
                                handleError("Duplicated entry in configuration file");
                                return false;
                            }
                            while (ln[++i] == ' ');
                            if (ln[i] == '\"') 
                            { 
                                i += 1;
                            }
                            int beg = i;
                            while (ln[++i] != ':');
                            int end = (ln[i-1] == '\"') ? i-1 : i;
                            String hostname = ln.Substring(beg, end-beg);
                            int port = Int32.Parse(ln.Substring(i+1));
                            storages[sid] = createStorage(this, sid);
                            storages[sid].open(hostname, port);
                        }
                    }                    
                    return true;
                } 
                finally 
                { 
                    reader.Close();
                }
            }
            catch (System.IO.IOException x)
            {
                handleException(x);
                return false;
            }		
        }
		
        /// <summary> Close database connection
        /// </summary>
        public virtual void Close()
        {
            for (int i = storages.Length; --i >= 0; )
            {
                storages[i].close();
            }
            CacheManager.removeFromCache(this);
        }
		
         /// <summary> Get root of the specified storage
        /// </summary>
        /// <param name="sid">storage identifier
        /// 
        /// </param>
        public virtual Object getRoot(int sid)
        {
            return storages[sid].Root;
        }
		
        /// <summary> Get size of the database file of the storage
        /// </summary>
        /// <param name="sid">storage identifier
        /// 
        /// </param>
        public virtual long getSize(int sid)
        {
            return storages[sid].Size;
        }
		
        /// <summary> Set root object for the specified storage in the database
        /// </summary>
        /// <param name="sid">storage identifier 
        /// </param>
        /// <param name="root">root object
        /// 
        /// </param>
        public virtual void  setRoot(int sid, Object root)
        {
            storages[sid].Root = (Persistent) root;
        }
		
        /// <summary> Associate persistent capable object with particular storage.
        /// Object should not yet be persistent (not belongs to soe storage)
        /// </summary>
        /// <param name="obj">persistent capable object
        /// </param>
        /// <param name="sid">storage identifier      
        /// 
        /// </param>
        public virtual void  attach(Object obj, int sid)
        {
            ObjectHeader p = ((Persistent) obj)._hdr();
            System.Diagnostics.Debug.Assert(p.opid == 0, "persistent object can't change its location");
            p.storage = storages[sid];
        }
		
       /// <summary> Parse oid into number and get that id from the first storage
        /// </summary>
        public virtual Persistent getObject(String oid)
        {
            try
            {
                int id = Int32.Parse(oid);
                return getObject(0, id);
            }
            catch (Exception e)
            {
                return null;
            }
        }
		
        /// <summary> Get object from the storage by OID
        /// </summary>
        /// <param name="storageId">storage
        /// </param>
        /// <param name="oid">object identifer 
        /// </param>
        /// <returns>fethed object or <code>null</code> if not found
        /// 
        /// </returns>
        public virtual Persistent getObject(int storageId, int oid)
        {
            return storages[storageId].getObject(oid);
        }
		
        /// <summary> Register event to be signaled on object modification
        /// </summary>
        /// <param name="obj">inspected object
        /// </param>
        /// <param name="e">event which will be signalled when object is modified by some other client
        /// 
        /// </param>
        public virtual void  notifyOnModification(Object obj, CondEvent e)
        {
            lock(this)
            {
                ObjectHeader p = ((Persistent) obj)._hdr();
                if (e == null)
                {
                    p.metaobject.notifyOnModification(p, false);
                    notifications.remove(p);
                }
                else
                {
                    p.metaobject.notifyOnModification(p, true);
                    notifications.put(p, e);
                }
            }
        }
		
        /// <summary> Register event to be signaled when transaction is aborted by server
        /// </summary>
        public virtual void  notifyOnTransactionAbort(CondEvent e)
        {
            lock(this)
            {
                onTransactionAbortEvent = e;
            }
        }
		
        /// <summary> Error handler. Override this method in derived class to provide 
        /// applciation specified error handling policy.
        /// </summary>
        /// <param name="text">error message
        /// 
        /// </param>
        protected internal virtual void  handleError(String text)
        {
            Console.Error.WriteLine(text);
            throw new ApplicationException("Application terminated due to critical error");
        }
		
        /// <summary> Error handler. Override this method in derived class to provide 
        /// applciation specified error handling policy.
        /// </summary>
        /// <param name="x">IOException
        /// 
        /// </param>
        protected internal virtual void  handleException(System.IO.IOException x)
        {
            Console.Error.Write(x.StackTrace);
            throw new ApplicationException("Application terminated due to critical error");
        }
		
        /// <summary> Create storage
        /// </summary>
        /// <param name="db">database
        /// </param>
        /// <param name="id">storage idenrtifier
        /// </param>
        /// <returns>created storage
        /// 
        /// </returns>
        protected internal virtual Storage createStorage(Database db, int id)
        {
            return new Storage(db, id);
        }
		
        /// <summary> Handler of client server connection failure. This method is called
        /// when connection with server is broken.
        /// Override this method in derived class to provide 
        /// applciation specified error handling policy.
        /// </summary>
        /// <param name="sid">storage identifier 
        /// 
        /// </param>
        protected internal virtual void  disconnected(int sid)
        {
            throw new SessionDisconnectedError("Server " + sid + " disconnect client session");
        }
		
        protected internal virtual void  abortTransaction()
        {
            for (int i = storages.Length; --i >= 0; )
            {
                storages[i].abortTransaction();
            }
            if (onTransactionAbortEvent != null)
            {
                onTransactionAbortEvent.signal();
            }
        }
    }
}