//-< IdentityHashtable.java >----------------------------------------*--------*
// GOODS                      Version 2.02       (c) 1998  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
// Java Program Interface                                            *  /  \  *
//                          Created:     18-Nov-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 18-Nov-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Hashtable based on objects identity (hash code returned by Object.hashCode)
//-------------------------------------------------------------------*--------*
namespace Goods
{
    using System;
    using System.Runtime.InteropServices;
	
    /// <summary> IdentityHashtable collision list.
    /// </summary>
    class IdentityHashtableEntry
    {
        internal Object key;
        internal Object val;
        internal IdentityHashtableEntry next;
    }
	
    /// <summary> This class implements a IdentityHashtable, which maps key object to some
    /// other object.
    /// An instance of <code>IdentityHashtable</code> has two parameters that 
    /// affect its efficiency: its <i>capacity</i> and its <i>load 
    /// factor</i>. The load factor should be between 0.0 and 1.0. When 
    /// the number of entries in the IdentityHashtable exceeds the product of the 
    /// load factor and the current capacity, the capacity is increased by 
    /// calling the <code>rehash</code> method. Larger load factors use 
    /// memory more efficiently, at the expense of larger expected time 
    /// per lookup. 
    /// 
    /// If many entries are to be made into a <code>IdentityHashtable</code>, 
    /// creating it with a sufficiently large capacity may allow the 
    /// entries to be inserted more efficiently than letting it perform 
    /// automatic rehashing as needed to grow the table. 
    /// </summary>
    public class IdentityHashtable
    {
        /// <summary> The hash table data.
        /// </summary>
        [NonSerialized()]
        private IdentityHashtableEntry[] table;
		
        /// <summary> The total number of entries in the hash table.
        /// </summary>
        [NonSerialized()]
        private int count;
		
        /// <summary> Rehashes the table when count exceeds this threshold.
        /// </summary>
        private int threshold;
		
        /// <summary> The load factor for the IdentityHashtable.
        /// </summary>
        private float loadFactor;
		
        /// <summary> Constructs a new, empty IdentityHashtable with the specified initial 
        /// capacity and the specified load factor. 
        ///
        /// </summary>
        /// <param name="initialCapacity">  the initial capacity of the Hashtable.
        /// </param>
        /// <param name="loadFactor">       a number between 0.0 and 1.0.
        /// </param>
        public IdentityHashtable(int initialCapacity, float loadFactor)
        {
            if ((initialCapacity <= 0) || (loadFactor <= 0.0))
            {
                throw new ArgumentException();
            }
            this.loadFactor = loadFactor;
            table = new IdentityHashtableEntry[initialCapacity];
            threshold = (int) (initialCapacity * loadFactor);
        }
		
        /// <summary> Constructs a new, empty IdentityHashtable with the specified initial 
        /// capacity and default load factor.
        ///
        /// </summary>
        /// <param name="initialCapacity">  the initial capacity of the IdentityHashtable.
        /// 
        /// </param>
        public IdentityHashtable(int initialCapacity):this(initialCapacity, 0.75f)
        {
        }
		
        /// <summary> Constructs a new, empty IdentityHashtable with a default capacity and load
        /// factor. 
        /// </summary>
        public IdentityHashtable():this(101, 0.75f)
        {
        }
		
        /// <summary> Returns the number of keys in this IdentityHashtable.
        ///
        /// </summary>
        /// <returns> the number of keys in this IdentityHashtable.
        /// 
        /// </returns>
        public virtual int size()
        {
            return count;
        }
		
        /// <summary> Tests if this IdentityHashtable maps no keys to values.
        ///
        /// </summary>
        /// <returns> <code>true</code> if this Hashtable maps no keys to values;
        /// <code>false</code> otherwise.
        /// 
        /// </returns>
        public virtual bool isEmpty()
        {
            return count == 0;
        }
		
        /// <summary> Tests if some key maps into the specified value in this IdentityHashtable.
        /// This operation is more expensive than the <code>containsKey</code>
        /// method.
        ///
        /// </summary>
        /// <param name="val">  a value to search for.
        /// </param>
        /// <returns>    <code>true</code> if some key maps to the
        /// <code>value</code> argument in this IdentityHashtable;
        /// <code>false</code> otherwise.
        /// 
        /// </returns>
        public virtual bool contains(Object val)
        {
            lock(this)
            {
                IdentityHashtableEntry[] tab = table;
                for (int i = tab.Length; i-- > 0; )
                {
                    for (IdentityHashtableEntry e = tab[i]; e != null; e = e.next)
                    {
                        if (e.val == val)
                        {
                            return true;
                        }
                    }
                }
                return false;
            }
        }
		
        /// <summary> Tests if the specified object is a key in this IdentityHashtable.
        /// 
        /// </summary>
        /// <param name="key">  possible key.
        /// </param>
        /// <returns> <code>true</code> if the specified object is a key in this
        /// IdentityHashtable; <code>false</code> otherwise.
        /// 
        /// </returns>
        public virtual bool containsKey(Object key)
        {
            lock(this)
            {
                IdentityHashtableEntry[] tab = table;
                int hash = key.GetHashCode();
                int index = (hash & 0x7FFFFFFF) % tab.Length;
                for (IdentityHashtableEntry e = tab[index]; e != null; e = e.next)
                {
                    if (e.key == key)
                    {
                        return true;
                    }
                }
                return false;
            }
        }
		
        /// <summary> Returns the value to which the specified key is mapped in this 
        /// IdentityHashtable.
        ///
        /// </summary>
        /// <param name="key">  a key in the IdentityHashtable.
        /// </param>
        /// <returns> the value to which the key is mapped in this IdentityHashtable;
        /// <code>null</code> if the key is not mapped to any value in
        /// this IdentityHashtable.
        /// 
        /// </returns>
        public virtual Object get(Object key)
        {
            lock(this)
            {
                IdentityHashtableEntry[] tab = table;
                int hash = key.GetHashCode();
                int index = (hash & 0x7FFFFFFF) % tab.Length;
                for (IdentityHashtableEntry e = tab[index]; e != null; e = e.next)
                {
                    if (e.key == key)
                    {
                        return e.val;
                    }
                }
                return null;
            }
        }
		
        /// <summary> Rehashes the contents of the IdentityHashtable into a IdentityHashtable 
        /// with a larger capacity. This method is called automatically when the 
        /// number of keys in the IdentityHashtable exceeds this IdentityHashtable's 
        /// capacity and load factor. 
        ///
        /// </summary>
        protected internal virtual void  rehash()
        {
            int oldCapacity = table.Length;
            IdentityHashtableEntry[] oldTable = table;
			
            int newCapacity = oldCapacity * 2 + 1;
            IdentityHashtableEntry[] newTable = new IdentityHashtableEntry[newCapacity];
			
            threshold = (int) (newCapacity * loadFactor);
            table = newTable;
			
            for (int i = oldCapacity; i-- > 0; )
            {
                for (IdentityHashtableEntry old = oldTable[i]; old != null; )
                {
                    IdentityHashtableEntry e = old;
                    old = old.next;
					
                    int index = (e.key.GetHashCode() & 0x7FFFFFFF) % newCapacity;
                    e.next = newTable[index];
                    newTable[index] = e;
                }
            }
        }
		
        /// <summary> Maps the specified <code>key</code> to the specified 
        /// <code>value</code> in this IdentityHashtable. Neither the key nor the 
        /// value can be <code>null</code>. 
        /// The value can be retrieved by calling the <code>get</code> method 
        /// with a key that is equal to the original key. 
        ///
        /// </summary>
        /// <param name="key">    the IdentityHashtable key.
        /// </param>
        /// <param name="val">  the value.
        /// </param>
        /// <returns>    the previous value of the specified key in this Hashtable,
        /// or <code>null</code> if it did not have one.
        /// 
        /// </returns>
        public virtual Object put(Object key, Object val)
        {
            lock(this)
            {
                // Makes sure the key is not already in the IdentityHashtable.
                IdentityHashtableEntry[] tab = table;
                int hash = key.GetHashCode();
                int index = (hash & 0x7FFFFFFF) % tab.Length;
                for (IdentityHashtableEntry e = tab[index]; e != null; e = e.next)
                {
                    if (e.key == key)
                    {
                        Object old = e.val;
                        e.val = val;
                        return old;
                    }
                }
				
                if (count >= threshold)
                {
                    // Rehash the table if the threshold is exceeded
                    rehash();
                    return put(key, val);
                }
				
                // Creates the new entry.
                IdentityHashtableEntry e2 = new IdentityHashtableEntry();
                e2.key = key;
                e2.val = val;
                e2.next = tab[index];
                tab[index] = e2;
                count++;
                return null;
            }
        }
		
        /// <summary> Removes the key (and its corresponding value) from this 
        /// IdentityHashtable. This method does nothing if the key is not in the 
        /// IdentityHashtable.
        ///
        /// </summary>
        /// <param name="key">  the key that needs to be removed.
        /// </param>
        /// <returns> the value to which the key had been mapped in this Hashtable,
        /// or <code>null</code> if the key did not have a mapping.
        /// 
        /// </returns>
        public virtual Object remove(Object key)
        {
            lock(this)
            {
                IdentityHashtableEntry[] tab = table;
                int hash = key.GetHashCode();
                int index = (hash & 0x7FFFFFFF) % tab.Length;
                for (IdentityHashtableEntry e = tab[index], prev = null; e != null; prev = e, e = e.next)
                {
                    if (e.key == key)
                    {
                        if (prev != null)
                        {
                            prev.next = e.next;
                        }
                        else
                        {
                            tab[index] = e.next;
                        }
                        count--;
                        return e.val;
                    }
                }
                return null;
            }
        }
		
        /// <summary> Clears this IdentityHashtable so that it contains no keys. 
        /// </summary>
        public virtual void  clear()
        {
            lock(this)
            {
                IdentityHashtableEntry[] tab = table;
                for (int index = tab.Length; --index >= 0; )
                {
                    tab[index] = null;
                }
                count = 0;
            }
        }
    }
}