//-< TestThreadPool.java >-------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     09-Sep-2001  K.A. Knizhnik  * / [] \ *
//                          Last update: 09-Oct-2001  K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Test for per-thread tranasctions
//-------------------------------------------------------------------*--------*

import goodsjpi.*;
import goodslib.*;

class Root extends Persistent { 
    ArrayOfObject films;

    Root() { 
	films = new ArrayOfObject(0);
    }

    void truncate() { 
	films.changeSize(0);
    }

    Film addFilm() {
	Film film = new Film(films.length());
	films.push(film);
	return film;
    }
}

class Film extends Persistent { 
    static final int BLOCK_SIZE = 256;

    int        id;
    ArrayOfInt data;

    void record(int id) { 
	data.insert(data.length(), BLOCK_SIZE, id);
    }

    Film(int id) { 
	this.id = id;
	data = new ArrayOfInt(0);
    }
}
    

class Transaction extends Thread { 
    static final int N_RECORDERS = 4;
    static final int N_FILMS = 32;

    int            id;
    CondEvent      completed;
    Database       db;
    Film           film;
    Root           root;
    ThreadPool     pool;
    RecorderThread thread;

    Transaction(ThreadPool pool, int id) { 
	this.pool = pool;
	this.id = id;
	completed = new CondEvent();
	db = new Database();
	boolean success = db.open("testtrans.cfg");	
	Assert.that(success);
    }

    public void reschedule(RecorderThread thread) { 
	int nextThreadId = 0;
	if (thread != null) { 
	    nextThreadId = thread.id + 1;
	    pool.releaseThread(thread);
	} 
	if (nextThreadId < N_RECORDERS) { 	    
	    pool.getThread(this, nextThreadId);
	} else { 
	    completed.signal();
	}
    }

    public void run() { 
	int i;
	db.attach();
	root = (Root)db.getRoot();
	for (i = 0; i < N_FILMS; i++) { 
	    BasicMetaobject.beginNestedTransaction();
	    film = root.addFilm();
	    reschedule(null);
	    completed.waitSignal();
	    BasicMetaobject.endNestedTransaction();
	}
	db.close();
    }
}


class RecorderThread extends Thread { 
    RecorderThread next;
    Transaction    trans;
    int            id;
    CondEvent      event = new CondEvent();
    boolean        running = true;

    void assign(Transaction trans, int id) { 
	this.trans = trans;
	this.id = id;
	event.signal();
    }

    void record() { 
	trans.db.attach();
	trans.film.record((trans.film.id*TestThreadPool.N_TRANSACTIONS 
			   + trans.id)*Transaction.N_RECORDERS + id);
    }

    void close() { 
	running = false;
	event.signal();
    }

    public void run() { 
	try { 
	    while (true) { 
		event.waitSignal();
		if (!running) { 
		    return;
		}
		record();
		trans.reschedule(this);
	    }
	} catch (Exception x) {}
    }
}

class ThreadPool { 
    RecorderThread threads;
    int            nThreads;
    
    synchronized RecorderThread getThread(Transaction trans, int id) { 
	if (threads == null) { 
	    RecorderThread recorder = new RecorderThread();
	    threads = recorder;
	    nThreads += 1;
	    recorder.start();
	    
	}
	RecorderThread thread = threads;       
	threads = thread.next;
	thread.assign(trans, id);
	return thread;
    }

    synchronized void releaseThread(RecorderThread thread) { 
	thread.next = threads;
	threads = thread;
    }

    void close() { 
	System.out.println("Threads in pool: " + nThreads);
	while (threads != null) { 
	    RecorderThread thread = threads;
	    threads = thread.next;
	    thread.close();
	}
    }
}

public class TestThreadPool { 
    static final int N_TRANSACTIONS = 16;
    static Transaction transactions[];

    static public void main(String args[]) throws Exception { 
	int i;
	transactions = new Transaction[N_TRANSACTIONS];
	Database db = new Database();
	boolean success = db.open("testtrans.cfg");
	Assert.that(success);
	Root root = (Root)db.getRoot();
	if (root == null) { 
	    root = new Root();
	    db.setRoot(root);
	} else { 
	    root.truncate();
	}
	ThreadPool pool = new ThreadPool();
	for (i = 0; i < N_TRANSACTIONS; i++) { 
	    transactions[i] = new Transaction(pool, i);
	    transactions[i].start();
	}
	for (i = 0; i < N_TRANSACTIONS; i++) { 
	    transactions[i].join();
	}
	pool.close();
	Assert.that(root.films.length() == N_TRANSACTIONS*Transaction.N_FILMS);
	int nRecordedFilmes[] = new int[N_TRANSACTIONS];
	for (i = 0; i < N_TRANSACTIONS*Transaction.N_FILMS; i++) { 
	    Film film = (Film)root.films.getAt(i);
	    int id = film.data.getAt(0);
	    int filmId = id / (N_TRANSACTIONS*Transaction.N_RECORDERS);
	    Assert.that(filmId == i);
	    int transId = id / Transaction.N_RECORDERS % N_TRANSACTIONS;
	    nRecordedFilmes[transId] += 1;
	    for (int j = 0; j < Transaction.N_RECORDERS; j++) { 
		for (int k = 0; k < Film.BLOCK_SIZE; k++) { 
		    id = film.data.getAt(j*Film.BLOCK_SIZE+k);
		    Assert.that(id / (N_TRANSACTIONS*Transaction.N_RECORDERS) == filmId);
		    Assert.that(id / Transaction.N_RECORDERS % N_TRANSACTIONS == transId);
		    Assert.that(id % Transaction.N_RECORDERS == j);
		}
	    }
	}
	for (i = 0; i < N_TRANSACTIONS; i++) { 
	    Assert.that(nRecordedFilmes[i] == Transaction.N_FILMS);
	}
	db.close();
    }
}


