package goodsjpi;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.tree.*;
import java.util.Vector;
import java.util.Enumeration;


/**
 * This is a simple GOODS browser.
 * It builds dynamically a tree 
 * to inspect objects in the database.
 * 
 * @author Guillaume BINET <guillaume.binet@hotguard.com>
 */
public class GoodsBrowser extends JPanel implements TreeWillExpandListener
{
   Database db;
   Persistent root;
   JTree tree;
   DefaultMutableTreeNode rootNode;
   DefaultTreeModel treeModel;
   Server server = null;
   JFrame mainWindow;


   public GoodsBrowser(String config, boolean startServer)
   {
      super(new BorderLayout());
      db = new Database();


      if( startServer )
      {
         server = new Server(config);
         server.start();
      }

      if( db.open(config) )
      {
         root = (Persistent) db.getRoot();

         if( root == null )
         {
            System.err.println("Can't browse an empty database");
            System.exit(-1);
         }
      }

      mainWindow = new JFrame();
      mainWindow.addWindowListener(
                                  new WindowAdapter () 
                                  {
                                     public void windowClosing (WindowEvent e) 
                                     {
                                        db.close();
                                        if( server != null )
                                        {
                                           server.stop();
                                        }
                                        System.exit(0);
                                     }
                                  });


      mainWindow.getContentPane().add(new JScrollPane (this));

      rootNode = new DefaultMutableTreeNode(new PersistentPresenter(root, "ROOT", "Root") );

      treeModel = new DefaultTreeModel(rootNode);

      tree = new JTree(treeModel);
      tree.addTreeWillExpandListener(this);
      tree.setCellRenderer(new TreeRenderer());
      ToolTipManager.sharedInstance().registerComponent(tree);

      add(tree, BorderLayout.CENTER);

      try
      {
         treeWillExpand(new TreeExpansionEvent(tree,new TreePath(rootNode.getPath())));
      }
      catch( Exception e )
      {
      }

      tree.expandPath(new TreePath(rootNode.getPath()));
      mainWindow.pack();
      mainWindow.setVisible(true);
   }

   /**
    * Invoked whenever a node in the tree is about to be collapsed.
    */
   public void treeWillCollapse(TreeExpansionEvent event) throws ExpandVetoException 
   {
      DefaultMutableTreeNode node = (DefaultMutableTreeNode)event.getPath().getLastPathComponent();
      node.removeAllChildren();
      treeModel.nodeStructureChanged( node );
      treeModel.insertNodeInto(new DefaultMutableTreeNode("dummy"),node,0);
   }

   private DefaultMutableTreeNode addNode(Object underlying, DefaultMutableTreeNode parent)
   {
      DefaultMutableTreeNode t = new DefaultMutableTreeNode(underlying);
      treeModel.insertNodeInto(t , parent, parent.getChildCount());
      return t;
   }


   /**
    * initialize a node with the given field
    * 
    * @param obj    the object that contains the field
    * @param fields the field descriptor
    * @param node   where you want to put it
    */
   private void dumpFieldsToNode(Persistent obj, FieldDescriptor fields, DefaultMutableTreeNode node)
   {
      Persistent.defaultMetaobject.preDaemon(obj, 0);
      try
      {

         FieldDescriptor current = fields;
         while( current != null )
         {


            Object next = current.field.get(obj);
            String name = current.field.getName();

            switch( current.appType )
            {
               case FieldDescriptor.t_boolean:
               case FieldDescriptor.t_byte:
               case FieldDescriptor.t_char:
               case FieldDescriptor.t_date:
               case FieldDescriptor.t_double:
               case FieldDescriptor.t_float:
               case FieldDescriptor.t_integer:
               case FieldDescriptor.t_long:
               case FieldDescriptor.t_short:
               case FieldDescriptor.t_string:
                  addNode(name + " = " + next, node);
                  break;

               case FieldDescriptor.t_reference:
                  if( next != null )
                  {

                     if( next.getClass().isArray() )
                     {
                        Persistent nextOnes[] = (Persistent[])next;
                        for( int i=0;i<nextOnes.length;i++ )
                        {
                           if( nextOnes[i] != null )
                           {
                              addNode("bidon", addNode(new PersistentPresenter(nextOnes[i], name + "[" + i + "]", buildTooltip(current)), node));
                           }
                           else
                           {
                              addNode(name + "[" + i + "] = NULL", node);
                           }
                        }
                     }
                     else
                     {
                        Persistent nextOne = (Persistent)next;
                        addNode("bidon", addNode(new PersistentPresenter(nextOne, name, buildTooltip(current)), node));

                     }
                  }
                  else
                  {
                     addNode(name + " = NULL", node);
                  }


                  break;
            }


            current = current.next;
         }
      }
      catch( Exception error )
      {
         System.err.println("error");
         error.printStackTrace();
      }
      Persistent.defaultMetaobject.postDaemon(obj, 0, false);

   }

   /**
    * Build the tooltip for a node
    * 
    * @param current the field descriptor from which you want to create the tooltip
    * @return the tooltip
    */
   private String buildTooltip(FieldDescriptor current)
   {
      return "<HTML>" +

      "<font face=arial size=2 color=black>" +
      "Field.size = " + 
      "</font>" +
      "<font face=arial size=2 color=red>" +
      current.size+
      "</font> <br> </HTML>";


   }

   /**
    * Invoked whenever a node in the tree is about to be expanded.
    */
   public void treeWillExpand(TreeExpansionEvent event) throws ExpandVetoException 
   {
      DefaultMutableTreeNode node = (DefaultMutableTreeNode)event.getPath().getLastPathComponent();

      node.removeAllChildren();
      treeModel.nodeStructureChanged( node );

      Persistent obj = ((PersistentPresenter)node.getUserObject()).getPersistent();

      dumpFieldsToNode(obj, obj.desc.firstField, node);


   }

   public static void main(String [] argv)
   {
      if( argv.length != 1 && (argv.length != 2 || !"start".equalsIgnoreCase(argv[1])) )
      {
         System.err.println("Incorrect command line: usage java goodspji.GoodsBrowser <goodsconfigfile> [start]");
         System.exit(-2);
      }

      GoodsBrowser gb = new GoodsBrowser(argv[0], argv.length == 2);

   }

}

/**
 * An holder that helps to present a node
 * 
 * @author Guillaume BINET <guillaume.binet@hotguard.com>
 */
class PersistentPresenter
{
   private Persistent p;
   private String fieldName;
   private String tooltip;

   public PersistentPresenter(Persistent p, String fieldName, String tooltip)
   {
      this.p=p;
      this.fieldName=fieldName;
      this.tooltip=tooltip;
   }

   public String toString()
   {
      return fieldName + " = " + beautiful();
   }

   public Persistent getPersistent()
   {
      return p;
   }

   private String beautiful()
   {
      StringBuffer ret = new StringBuffer("");
      /*
      if( p instanceof ArrayOfByte )
      {
         ArrayOfByte array = (ArrayOfByte) p;
         for( int i=0;i<array.length();i++ )
         {
            int byteToPrint = array.getAt(i);

            if( byteToPrint<0 )
            {
               byteToPrint+=0xFF;
            }

            if( byteToPrint<16 )
            {
               ret.append("0");
            }

            ret.append((Integer.toHexString(byteToPrint)+" ").toUpperCase());
         }
      }
      else
      */
      ret.append(p.toString());
      return ret.toString();


   }

   public String getTooltip()
   {
      return tooltip;
   }

}

/**
 * A custom treerenderer to insert tooltips
 * 
 * @author Guillaume BINET <guillaume.binet@hotguard.com>
 */
class TreeRenderer extends DefaultTreeCellRenderer
{
   public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus)
   {

      //log.debug("get treecellrenderercomponent");
      super.getTreeCellRendererComponent(tree, value, sel,expanded, leaf, row,hasFocus);

      Object realValue = ((DefaultMutableTreeNode)value).getUserObject();

      //log.debug("value = " + realValue);
      if( realValue instanceof PersistentPresenter )
      {
         setToolTipText (((PersistentPresenter)realValue).getTooltip());
      }

      if( leaf == true )
      {
         //log.debug("autre");
         setToolTipText ("blahblah" );
      }

      return this;
   }
}






