package goodsjpi;

// Global imports
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.io.IOException;

/**
 * Implementation of PersistenceFactory which creates
 * goods specific persistent objects.
 * Use lines like
 *    persistence.dbname=goods
 *    persistence.roottype=my.root.object.type
 *    goodsjpi.meta=goodsjpi.PessimisticMetaobject
 * in the properties file
 * The global metaobject type may be defined with a 
 * line of the form goodsjpi.meta=<metaobject class>
 * Note that this feature should be used with care !
 * @author Joel Crisp
 */
public class GoodsFactory extends PersistenceFactory
{
    
    /**
     * Default constructor. Protected since the singleton is created
     * by the superclass
     */
    protected GoodsFactory() throws IOException
    {
        super();
        String defaultMetaObjectType=persistenceMap.getProperty("goodsjpi.meta");
        if(defaultMetaObjectType==null) return;
        Class metaObjectClazz=null;
        try
            {
                metaObjectClazz=Class.forName(defaultMetaObjectType);
            }
        catch(ClassNotFoundException cnfe)
            {
                System.err.println("Metaobject type "+defaultMetaObjectType+" not found.");
                cnfe.printStackTrace();
                return;
            }

        try
            {
                Persistent.defaultMetaobject=(Metaobject)metaObjectClazz.newInstance();
            }
        catch(InstantiationException ie)
            {
                System.err.println("Metaobject type "+defaultMetaObjectType+"can't be instantiated.");
                ie.printStackTrace();
            }
        catch(IllegalAccessException iae)
            {
                iae.printStackTrace();
            }
    }

    /**
     * Get a persistence instance compatible with the required class
     * The required class may be an interface type or a concrete type.
     * If it is a concrete type then the returned value will either be
     * an instance of that type or one of its subclasses, if it is an
     * interface type then the returned object must implement that interface
     * The mapping is looked up in the persistence.properties file
     * @param required the type of the class or interface for which a 
     *   persistent instance is required.
     */
    public final Object getImplementation(Class required)
    {
        Class implClass=getImplClass(required);
        try
            {
                Object result=implClass.newInstance();
                if(result instanceof Persistent)
                    {
                        Persistent persistent=(Persistent)result;
                        persistent.metaobject.postDaemon(persistent,Metaobject.CONSTRUCTOR,true);
                        return persistent;
                    }
                else return result;
            }
        catch(InstantiationException ie)
            {
                ie.printStackTrace();
                return null;
            }
        catch(IllegalAccessException iae)
            {
                iae.printStackTrace();
                return null;
            }
    }

    /**
     * Get a persistence instance compatible with the required class
     * The required class may be an interface type or a concrete type.
     * If it is a concrete type then the returned value will either be
     * an instance of that type or one of its subclasses, if it is an
     * interface type then the returned object must implement that interface
     * The mapping is looked up in the persistence.properties file
     * This version of the method calls the constructor whose signature
     * matches the required types
     * @param required the type of the class or interface for which a 
     *   persistent instance is required.
     * @param types array of parameter types for the constructor
     * @param params the arguments to the constructor (primitive types wrapped)
     */
    public final Object getImplementation(Class required,Class types[],Object params[])
    {
        Class implClass=getImplClass(required);
        try
            {
                Constructor constructor=implClass.getConstructor(types);

                Object result=constructor.newInstance(params);
                if(result instanceof Persistent)
                    {
                        Persistent persistent=(Persistent)result;
                        persistent.metaobject.postDaemon(persistent,Metaobject.CONSTRUCTOR,true);
                        return persistent;
                    }
                else return result;

            }
        catch(NoSuchMethodException nsme)
            {
                nsme.printStackTrace();
                return null;
            }
        catch(InstantiationException ie)
            {
                ie.printStackTrace();
                return null;
            }
        catch(IllegalAccessException iae)
            {
                iae.printStackTrace();
                return null;
            }

        catch(InvocationTargetException ite)
            {
                ite.printStackTrace();
                return null;
            }
    }

    /**
     * Retrieve or create the root object of the datastore
     * The root object type is specified by "persistence.roottype"
     * @return the root object of the datastore
     */
    public synchronized Object getRootImpl() throws IOException
    {
        if(root!=null) return root;
        Database db = new Database();
        if (db.open(persistenceMap.getProperty("persistence.dbname")+".cfg"))
            {
                root = db.getRoot();
                if (root == null)
                    {
                        String rootType=persistenceMap.getProperty("persistence.roottype");
                        try
                            {
                                Class rootClazz=Class.forName(rootType);
                                root = rootClazz.newInstance();
                            }
                        catch(ClassNotFoundException cnfe)
                            {
                                System.err.println("FATAL: Can't instantiate root object");
                                System.exit(-3);
                            }
                        catch(InstantiationException ie)
                            {
                                System.err.println("FATAL: Can't instantiate root object");
                                System.exit(-4);
                            }
                        catch(IllegalAccessException iae)
                            {
                                System.err.println("FATAL: Can't instantiate root object, no-arg constructor must be public");
                                System.exit(-5);
                            }
                        if(root instanceof Persistent)
                            {
                                Persistent persistent=(Persistent)root;
                                persistent.metaobject.postDaemon(persistent,Metaobject.CONSTRUCTOR,true);
                            }

                        db.setRoot(root);
                    }
                return root;
            }
        return null;
    }

}
