package goodslib;

import java.util.Map;
import java.util.Iterator;
import java.lang.reflect.InvocationTargetException;

/**
 * Creates a goodslib.QueryIterator to iterate over the
 * supplied Map.  A shallow copy of the map is created.  This means that this
 * iterator captures the state of the map as it was at the iterator's creation
 * time and the map is not kept up-to-date with the original.
 *
 * <p>This class is guaranteed to work correctly even if you have multiple
 * callers using the iterator at the same time.  Note however that if
 * subsequent calls to getNext during runtime are not marching sequentially
 * through the iterator, a large performance hit may be observed.</p>
 *
 * <p>No guarantees are made if the iterator is associated with a
 * <code>WeakHashMap</code> or any other <code>Map</code> that uses weak
 * references.</p>
 *
 * <p>This class is believed to be threadsafe.</p>
 *
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Alphatech, Inc. </p>
 * @author Gerald Dalley
 * @version 1.0
 */

public class MapQueryIterator
    implements QueryIterator {
  public String getKeyName() {
    return keyName;
  }

  /**
   * An exception will be thrown if the supplied map object does not implement
   * a public <code>clone()</code> method.
   *
   * @param map Map         The map over which this iterator will operate.  The
   *                        Map must supply a public <code>clone</code> method.
   * @param keyName String  Name of the key field.  All calls to
   *                        <code>getByKey</code> must use this key name.
   * @throws SecurityException
   * @throws NoSuchMethodException
   * @throws InvocationTargetException
   * @throws IllegalArgumentException
   * @throws IllegalAccessException
   */
  public MapQueryIterator(Map map, String keyName) throws SecurityException,
      NoSuchMethodException, InvocationTargetException,
      IllegalArgumentException, IllegalAccessException {
    if (map == null) {
      throw new NullPointerException(
          this.getClass().getName() + " must be associated with a non-null " +
          this.map.getClass().getName() + ".");
    }
    synchronized (map) {
      this.map = (Map)map.getClass().getMethod("clone", null).invoke(map, null);
    }
    this.keyName = keyName;
    this.prevIndex = 0;
  }

  /**
   * Returns the first object indexed by this iterator.  This moves the
   * internal iterator cached location to the second item.
   *
   * @return Object   First value in the map (according to the original
   *                  map's ordering scheme).
   */
  public synchronized Object getFirst() {
    nextItemIter = map.values().iterator();
    this.prevIndex = -1;
    this.advanceCurrIter();

    return currObject;
  }

  /**
   * Private method that advances the current iterator forward by one
   * position, updating all of the internal iterator cache fields.  This method
   * assumes that the <code>prevIndex</code>, <code>nextItemIter</code>,
   * and <code>currObject</code> members are all consistent on entry and does
   * not do any checking.
   */
  private synchronized void advanceCurrIter() {
    if (nextItemIter.hasNext()) {
      currObject = nextItemIter.next();
    }
    else {
      currObject = null;
    }
    this.prevIndex++;
  }

  /**
   * Returns the next object in the iteration list.  This method verifies first
   * that a reference to <code>prevObject</code> exists at iteration position
   * <code>prevIndex</code>.  If <code>prevObject</code> and
   * <code>prevIndex</code> do not agree or are otherwise invalid, an exception
   * is thrown.  Assuming no problems were encountered, the next object
   * in the iteration list is extracted and the internal iterator cache is
   * updated to optimize the next call to <code>getNext</code>.
   *
   * <p>If <code>getNext</code> is called with valid arguments, but the
   * previous object and index are not actually the most recently returned
   * values, this method will still function properly, but a time-consuming
   * search may be required to find the next item.</p>
   *
   * @param prevObject Object   reference returned by previous call of
   *                            <code>getNext</code> or <code>getFirst</code>.
   * @param prevIndex int       index of previous object (starting from 0 for
   *                            first object).
   * @return Object             next object or <code>null</code> if there are
   *                            no more objects
   * @throws IllegalArgumentException  Thrown if the object at
   *      <code>prevIndex</code> is not the same as <code>prevObject</code>
   *      (the actual references must match).
   */
  public Object getNext(Object prevObject, int prevIndex) {
    if (this.prevIndex == prevIndex) {
      if (prevObject == currObject) {
        // The passed-in prevObject is indeed the most recently used previous
        // object that we have cached.  Now we can just use the cached iterator.
        this.advanceCurrIter();
      }
      else {
        throw new IllegalArgumentException(
            "The item at location " + prevIndex + " does not match the " +
            "prevObject paramater supplied by the caller.");
      }

    }
    else if (prevIndex >= map.size()) {
      // The iterator has passed the end of the list of map entries.  Assuming
      // they didn't mess up prevObject, just keep returning null forever.
      if (prevObject == null) {
        // User didn't make any mistakes
        currObject = null;
      }
      else {
        throw new IllegalArgumentException(
            "The supplied previous index (" + prevIndex + ") is past the end " +
            "of iterator, but the caller claims that the previous object " +
            "returned by this iterator was not null.");
      }

    }
    else {
      // Either the caller messed up or this "iterator" is being used by
      // multiple clients that are running at different stages in the iteration.
      // Our cached iterator won't work, so we need to find and verify the
      // passed-in prevObject and prevIndex before fetching the next object.

      // Find the prevIndex item.  Note that we don't need to bother checking
      // nextItemIter.hasNext because we already verified that
      // (prevIndex < map.size())
      Iterator iter = map.values().iterator();
      Object o = null;
      for (int idx = 0; idx <= prevIndex; idx++) {
        o = iter.next();
      }

      if (o == prevObject) {
        // We found the specified location and verified that they object
        // that the caller thinks is at prevIndex is indeed there.  Time to
        // cache our new location and move on.
        synchronized (this) {
          this.nextItemIter = iter;
          this.prevIndex = prevIndex;
          this.currObject = o;
          this.advanceCurrIter();
        }
      } else {
        throw new IllegalArgumentException(
            "The item at location " + prevIndex + " does not match the " +
            "prevObject paramater supplied by the caller.");
      }
    }

    return currObject;
  }

  /**
   * To be consistent with the documentation for QueryInterface,
   * <code>getByKey</code> does not modify the state of the iterator.  This
   * means that calls to <code>getByKey</code> are independent of calls to
   * <code>getNext</code> and <code>getFirst</code> and will not impose
   * any performance hits on otherwise normal <code>getFirst</code> +
   * <code>getNext</code> iteration.
   *
   * @param keyName String
   * @param keyValue Object
   * @return Object
   */
  public Object getByKey(String keyName, Object keyValue) {
    if (this.keyName == null) {
      if (keyName != null) {
        throw new NoIndexException();
      }
    } else if (!this.keyName.equals(keyName)) {
      throw new NoIndexException();
    }
    return map.get(keyValue);
  }

  private Map map = null;
  private String keyName = null;
  private Iterator nextItemIter = null;
  private Object currObject = null;
  private int prevIndex = 0;
}
